import pickle as pk
import os
import io
from aws_s3_resource.s3_bucket import S3Bucket
from aws_s3_resource.s3_object import S3Object
from .setup import Setup
import time 


class EventManager:
    """ This class works on  determine events and save them """
    basic_events = ['tweet_delete_events', 'direct_message_mark_read_events',
                    'direct_message_indicate_typing_events']

    type_based_events = ['follow_events','mute_events','block_events']

    # Conditons to determine who did the event
    by_to_conditions = {
        'by_if_equal': lambda id1, id2: 'by' if id1 == id2 else 'to',
        'to_if_equal': lambda id1, id2: 'to' if id1 == id2 else 'by'
    }
    
    def __init__(self, storing_medium='local', events_folder=os.getcwd(), bucket_name = None):
        # Call setup function , define save_func according to storing_medium
        if storing_medium == 'local':
            self.events_folder = events_folder
            Setup(storing_medium=storing_medium, events_folder=events_folder)
            self.save_func = self.save2pickle

        elif storing_medium == 's3':
            self.bucket_name = bucket_name
            Setup(storing_medium=storing_medium, bucket_name=bucket_name)
            self.save_func = self.save2s3

    # Svae event to S3
    def save2s3(self, pickle_name, response):
        object_key  = f'{pickle_name}.pkl'
        logg_bytes = S3Object.download(self.bucket_name, object_key)

        with io.BytesIO(logg_bytes) as f:
            loggs = pk.load(f)

        action = {f'{pickle_name}' : response}

        # Add current action to log
        action['timestamp'] = int(time.time())
        loggs.append(action)

        data_bytes = pk.dumps(loggs)

        file_dict = {
            'file_name': object_key,
            'file_bytes': data_bytes
        }

        S3Object.upload(self.bucket_name, file_dict, with_random_prefix=False)
        return f'successful {pickle_name}'

    # Svae event to pickle file
    def save2pickle(self, pickle_name, response):
      with open(os.path.join(self.events_folder, f'{pickle_name}.pkl'), 'rb') as f:
          events = pk.load(f)

      events.append(response)

      with open(os.path.join(self.events_folder, f'{pickle_name}.pkl'), 'wb') as f:
          pk.dump(events, f)

      return f'successful {pickle_name}'

    def save_events(self, response):
        # Get events list
        keys_list = response.keys()

        # Get event name
        event_name = self.get_activity_name(keys_list)

        if event_name in EventManager.basic_events:
            pickle_name = event_name

        else:
            # If event depends on type 
            if event_name in EventManager.type_based_events:

                # Get event type
                event_type = response[f'{event_name}'][0]['type']

                # Test if the event done by user or to user 
                if event_type == 'follow':
                    follow_to = response['follow_events'][0]['target']['id']
                    user_id = response['for_user_id']
                    res = EventManager.by_to_conditions['to_if_equal'](user_id, follow_to)
                    pickle_name = f'follow_{res}_user_events'

                # If event is unfollow, mute, unmute, block, or unblock
                else:
                    pickle_name = f'{event_type}_by_user_events'

            elif event_name == 'direct_message_events':
                message_by = response["direct_message_events"][0]['message_create']['sender_id']
                user_id = response['for_user_id']
                res = EventManager.by_to_conditions['by_if_equal'](message_by, user_id)
                pickle_name = f'message_{res}_user_events'

            elif event_name == 'favorite_events':
                favorite_by = response['favorite_events'][0]['user']['id_str']
                user_id = response['for_user_id']
                res = EventManager.by_to_conditions['by_if_equal'](favorite_by, user_id)
                pickle_name = f'favorite_{res}_user_events'

            elif event_name == 'tweet_create_events':
                # Tweet create events maybe tweet (by user), retweet(to user,by user) , reply(to user,by user),mention (to user)
                if response['tweet_create_events'][0]['in_reply_to_status_id'] is None:
                   # Just to recocnize is it retwee or tweet by user
                   # If no it is tweet event
                    if "retweeted_status" in response['tweet_create_events'][0].keys():
                        retweeted_by = response['tweet_create_events'][0]['user']['id_str']
                        user_id = response['for_user_id']

                        res = EventManager.by_to_conditions['by_if_equal'](retweeted_by, user_id)
                        pickle_name = f'retweet_{res}_user_events'

                    else:
                        pickle_name = event_name

                else:

                    if response['tweet_create_events'][0]['in_reply_to_user_id_str'] == response['for_user_id']:
                        pickle_name = 'reply_to_user_events'

                    else:
                        if response['tweet_create_events'][0]['user']['id_str'] == response['for_user_id']:
                            pickle_name = 'reply_by_user_events'
                        else:
                            pickle_name = 'mention_to_user_events'


        return self.save_func(pickle_name, response)

    def get_activity_name(self, keys_list):
        for k in keys_list:
            if 'event' in k:
                return k


import json, time, datetime, tweepy, math, functools
from .twitter_api_preparation import TwitterAPIsList, TwitterAPI
from abc import ABCMeta, abstractmethod
from langdetect import detect


class APIsCursor:
    """ This class is responsible for changing the API when it reaches to the maximum
        limit
    """
    def __init__(self, waiting_period=0.5, maximum_period=900):
        
        self.apis = TwitterAPIsList()
        self.number_of_trials = 0
        self.number_of_apis = self.apis.number_of_apis
        self.waiting_period = waiting_period
        self.cur_waiting_period = waiting_period
        self.maximum_period = maximum_period
        self.current_api = -1
        
    def update(self, required_function=None):
        """ Check what is the next available API """
        
        if self.number_of_trials == self.number_of_apis:
            # All the apis exceed the rate limit
    
            # Reset number of trials
            self.number_of_trials = 0
            
            # Waiting
            self.cur_waiting_period *= 2
            print(f'Waiting for : {self.cur_waiting_period}')
            time.sleep(self.cur_waiting_period)
            
            if self.cur_waiting_period > self.maximum_period:
                self.cur_waiting_period = self.waiting_period
                
        else:
            self.number_of_trials += 1
            
            # Use the next api
            self.current_api = (self.current_api + 1) % self.number_of_apis
            print(f'Using the api number {self.current_api} for the request')
        
        api = self.apis[self.current_api]
        
        if required_function:
            # Return the required function from the active API
            return getattr(api, required_function)
        return api


class TweetsCollector(metaclass=ABCMeta):
    """ This is the parent class for any class to collect tweets """
    def __init__(self, search_params, collection_function):
        self.search_params = search_params
        
        
        self.tweets = []   # Empty list to keep the collected tweets
        self.collection_function = collection_function   # The required function to collect data
        self.api_cursor = APIsCursor()  # Cursor to change the API what hitting the limit
        
        self.tweets_cursor = None
        self.update_tweets_cursor()
        
        self.collection_done = False  # Flag to exit once collection is done
    
    def update_tweets_cursor(self):
        """ Get another api function cursor when hitting the limit """
        # Update max_id to collect tweets that are before the tweets 
        # that already is collected
        common_search_parms = {
            'max_id': None if len(self.tweets) == 0 else self.tweets[-1].id
        }

        collector = self.api_cursor.update(self.collection_function)
        self.tweets_cursor = tweepy.Cursor(
            collector, **{**common_search_parms, **self.search_params}).pages()
    

    def filter_out_repeated_tweets(self):
        """ This function will filter out any possible repeated tweet """
        print('Filtering ....')
        tweets_dict = {tw.id_str: tw for tw in self.tweets}

        if len(tweets_dict) != len(self.tweets):
            print('Number of repeated tweets is : ', len(self.tweets)- len(tweets_dict) )
            
            # Take the unique tweets only
            self.tweets = list(tweets_dict.values())
    
    @abstractmethod
    def collect(self):
        pass


class UserTimelineTweetsCollector(TweetsCollector):
    """ This class is to collect tweets according to a user id """
    
    def __init__(self, search_params):
        super().__init__(search_params, 'user_timeline') 
    
    def collect(self, number_of_tweets=math.inf):
        """ This function collects tweets and stops when the number of tweets
            reaches number_of_tweets
        """
        
        while not self.collection_done:
            try:
                page = next(self.tweets_cursor)
                
                for tw in page:
                    self.tweets.append(tw)
                    
                    if len(self.tweets) == number_of_tweets:
                        raise StopIteration
            
            except tweepy.TweepError:
                if TwitterAPI.is_rate_limit_error(e):
                    # Hitting the rate limit. Therefore, change the API
                    self.update_tweets_cursor()
                else:
                    # Unknown: it should be fixed once happening
                    raise e
            
            except StopIteration:
                # No More Tweets
                self.collection_done = True
                break


class QueryTweetsCollector(TweetsCollector):
    """ This class is to collect tweets according to a keyword """
    
    def __init__(self, search_params):
        super().__init__(search_params, 'search') 
    
    
    def collect(self, filter_on_english, number_of_tweets=math.inf):
        """ This function collects tweets and stops when the number of unique account
            reaches number_of_accounts
        """
        tweets_accounts = set()
        
        while not self.collection_done:
            try:
                page = next(self.tweets_cursor)
                
                for tw in page:
                    if filter_on_english:
                        try:
                            if detect(tw._json['full_text']) == 'en':
                                self.tweets.append(tw)

                        except:
                            pass
                    
                    if len(self.tweets) == number_of_tweets:
                        raise StopIteration
            
            except tweepy.TweepError as e:
                if TwitterAPI.is_rate_limit_error(e):
                    # Hitting the rate limit. Therefore, change the API
                    self.update_tweets_cursor()
                else:
                    # Unknown: it should be fixed once happening
                    raise e
            
            except StopIteration:
                # No More Tweets
                self.collection_done = True
                break


class FavoritedTweetsCollector(TweetsCollector):
    """ This class is to collect the favourited tweets by a user """
    
    def __init__(self, search_params):
        super().__init__(search_params, 'favorites') 
    
    def collect(self, number_of_tweets=math.inf):
        """ This function collects tweets and stops when the number of tweets
            reaches number_of_tweets
        """
        
        while not self.collection_done:
            try:
                page = next(self.tweets_cursor)
                
                for tw in page:
                    self.tweets.append(tw)
                    
                    if len(self.tweets) == number_of_tweets:
                        raise StopIteration
            
            except tweepy.TweepError as e:
                
                if TwitterAPI.is_rate_limit_error(e):
                    # Hitting the rate limit. Therefore, change the API
                    self.update_tweets_cursor()
                else:
                    # Unknown: it should be fixed once happening
                    raise e
                
            except StopIteration:
                # No More Tweets
                self.collection_done = True
                break


class DataCollector:
    """ This class is to collect any type of data from twitter except tweets(Refer to tweets
        collector) 
    """
    def __init__(self, required_function):
        self.api_cursor = APIsCursor()  # Cursor to change the API what hitting the limit
        self.required_function = required_function  # The function to be used to get the data
        self.function = None
        self.update_api_function()
    
    def update_api_function(self):
        self.function = self.api_cursor.update(self.required_function)
    
    def get(self, *args, **kwargs):
        while True:
            try:
                data = self.function(*args, **kwargs)
                break
            except tweepy.TweepError as e:
                if TwitterAPI.is_rate_limit_error(e):
                    self.update_api_function()
                else:
                    raise e
        return data
    
    @staticmethod
    def get_user_collector():
        return DataCollector('get_user')
    
    @staticmethod
    def get_status_collector():
        return DataCollector('get_status')
    
    @staticmethod
    def show_friendship_collector():
        return DataCollector('show_friendship')
    
    @staticmethod
    def trends_available_collector():
        return DataCollector('trends_available')
    
    @staticmethod
    def trends_place_collector():
        return DataCollector('trends_place')



import tweepy
import pandas as pd
import numpy as np
import pickle as pk
import json
from .twitter_api_preparation import TwitterAPI
from .twitter_data_collection import TweetsCollector, DataCollector
from abc import ABCMeta, abstractmethod
from langdetect import detect


class TrendsCollector(metaclass=ABCMeta):
    """ This class is an abstract class to collect trends topics and trends places """
    def __init__(self, columns_to_be_dropped=[], col_names_to_be_changed=dict()):
        self.data_df = None
        self.columns_to_be_dropped = columns_to_be_dropped
        self.col_names_to_be_changed = col_names_to_be_changed
    
    @abstractmethod
    def request_data(self):
        pass
    
    def fix_some_columns(self):
        pass
    
    def rename_some_columns(self):
        self.data_df.rename(self.col_names_to_be_changed, inplace=True, axis=1)
    
    def drop_some_columns(self):
        self.data_df.drop(self.columns_to_be_dropped, inplace=True, axis=1)
    
    def add_some_columns(self):
        pass


class TrendsPlaces(TrendsCollector):
    """ This class is responsible for getting the trends places from the twtiiter API """
    
    def __init__(self):
        super().__init__(['url', 'parentid', 'place_code'], 
                         {
                             'name': 'place_name',
                             'countryCode': 'country_code'
                         })
        
    def get_data_df(self, number_of_places):
        self.request_data(number_of_places)
        self.fix_some_columns()
        self.rename_some_columns()
        self.drop_some_columns()
        self.add_some_columns()
        
        return self.data_df
    
    def request_data(self, number_of_places):
        """ Get the list of trends places from Twitter API """
        ta = DataCollector.trends_available_collector()
        trends_places = ta.get()
        self.data_df = pd.DataFrame(trends_places[:number_of_places])
    
    def fix_country_code_col(self):
        """ Put each None in countryCode as Worldwide """
        self.data_df['countryCode'] = self.data_df['countryCode'].apply(
                                        lambda x: 'Worldwide' if x is None else x)
        
    def fix_country_col(self):
        """ Put each None in country as Worldwide """
        self.data_df['country'] = self.data_df['country'].apply(
                                        lambda x: 'Worldwide' if x is '' else x)
        
    def fix_place_type_col(self):
        """ Convert the place_type dict into two columns in the dataframe """
        
        pt = self.data_df.pop('placeType')
        pt_df = pd.DataFrame(pt.values.tolist(), index = self.data_df.index)
        pt_df.columns = ['place_code', 'place_type']
        self.data_df = self.data_df.join(pt_df)
    
    def fix_some_columns(self):
        self.fix_country_code_col()
        self.fix_country_col()
        self.fix_place_type_col()
    
    def get_woeids_list(self):
        return self.data_df.woeid.values.tolist()
    
    def add_woeid_places_coordinates(self):
        """ Get actual coordinates for each place using Yahoo API """
        
        woeids_list = self.get_woeids_list()
        coords_list = []
        for woeid in woeids_list:
            coord = {'long': 0, 'lat': 0} # yahoo_api.get_lng_lat(woeid)  # TODO
            coords_list.append(coord)
        
        coords_df = pd.DataFrame(coords_list)
        self.data_df = self.data_df.join(coords_df)
    
    def add_some_columns(self):
        self.add_woeid_places_coordinates()


class TrendsTopics(TrendsCollector):
    """ This class is responsible for getting the trends topics from the twtiiter API """
    def __init__(self, woeids_list, filter_on_english=True):
        super().__init__(['promoted_content', 'query'],
                       {
                           'url': 'trend_twitter_url',
                           'name': 'trend_topic'
                       })
        self.woeids_list = woeids_list
        self.filter_on_english = filter_on_english
    
    def get_data_df(self):
        self.request_data()
        self.fix_some_columns()
        self.rename_some_columns()
        self.drop_some_columns()
        self.add_some_columns()
        
        return self.data_df
    
    def request_data(self):
        """ Get the trends for each place """
        tp = DataCollector.trends_place_collector()
        trends_topics = list()

        for woeid in self.woeids_list:
            # Get the trends
            trends = tp.get(woeid)
            
            # Get trends place id only
            location = trends[0]['locations'][0]
            location.pop('name')
            
            # Merge id with the trends
            if self.filter_on_english:
                trends = [{**trnd, **location} for trnd in trends[0]['trends'] if detect(trnd['name']) == 'en']

            else:
                trends = [{**trnd, **location} for trnd in trends[0]['trends']]
            
            # Add to the topics list
            trends_topics.extend(trends)
            
        # Create a dataframe for the whole world trends
        self.data_df = pd.DataFrame(trends_topics)
                
        return self.data_df


class Trends:
    """ This class is responsible for getting the trends places and topics from 
        the twtiiter API 
    """
    def __init__(self):
        self.trends_places_df = None
        self.trends_topics_df = None
        self.trends_df = None
    
    def get_current_trends_df(self, filter_on_english, number_of_places=10):
        # Get trends places dataframe
        print('Get trends places .. ')
        tr_pl = TrendsPlaces()
        self.trends_places_df = tr_pl.get_data_df(number_of_places)
        
        # Get trends topics dataframe
        print('Get trends topics .. ')
        tr_tp = TrendsTopics(tr_pl.get_woeids_list(), filter_on_english=filter_on_english)
        self.trends_topics_df = tr_tp.get_data_df()
        
        print('Merging')
        # Create the full trends dataframe
        self.trends_df = self.trends_topics_df.merge(self.trends_places_df, how='left', 
                                                     left_on='woeid', right_on='woeid')
        
        Trends.convert_nan_to_none(self.trends_df, 'tweet_volume')
        return self.trends_df
        
    @staticmethod
    def convert_nan_to_none(df, key):
        df[key] = df[key].where(pd.notnull(df[key]), None)


import os
from abc import ABCMeta, abstractmethod
from .twitter_account import TwitterAccountProxy
import random 
import json
import threading
import time


class TwitterBot:
    def __init__(self,  bot_crediential_file,
        strategy_time_interval_range, action_time_interval_range,
        strategy_class, strategy_params): 
    
        self.bot_account = TwitterAccountProxy(bot_crediential_file)
        self.strategy_time_interval = random.randint(*strategy_time_interval_range)
        self.action_time_interval = random.randint(*action_time_interval_range)
        self.strategy_class = strategy_class
        self.strategy_params = strategy_params
        
    
    def go(self):
        strategy = self.strategy_class(**self.strategy_params)
        actions = strategy.start()
        
        for action in actions:
            # Do action
            self.bot_account.do(action)
            time.sleep(self.action_time_interval * 60)
        
        threading.Timer(self.strategy_time_interval * 3600 , self.go).start()

        


import tweepy
import random
import json
import time
import pickle as pk
from .twitter_action import TwitterAction
from botocore.exceptions import ClientError


class TwitterAccountProxy(tweepy.API):

    def __init__(self,
        token_file='twitter_tokens.json', wait_on_rate_limit=False,
        timeout=10, retry_count=3):


        # Dictionary contains twitter application keys
        with open(token_file, 'r') as f:
            self.token = json.load(f)
                
        auth = tweepy.OAuthHandler(self.token['consumer_key'],
                                   self.token['consumer_secret'])
        
        auth.set_access_token(self.token['access_token'],
                              self.token['access_token_secret'])
        
        # Call tweepy.API to create api object
        super().__init__(auth, wait_on_rate_limit=wait_on_rate_limit,
                         timeout=timeout, retry_count=retry_count)
        
    @staticmethod
    def follow_user(self, user_screen_name):
        # user_screen_name could be either screen name or user id
        self.create_friendship(user_screen_name)


    def unfollow_user(self, user_screen_name):
        # user_screen_name could be either screen name or user id
        self.destroy_friendship(user_screen_name)

    def block_user (self,user_screen_name):
        # user_screen_name could be either screen name or user id
        self.create_block (user_screen_name)

    def unblock_user(self,user_screen_name):
        # user_screen_name could be either screen name or user id
        self.destroy_block(user_screen_name)

    def mute_user(self,user_screen_name):
        # user_screen_name could be either screen name or user id
        self.create_mute(user_screen_name)

    def unmute_user(self,user_screen_name):
        # user_screen_name could be either screen name or user id
        self.destroy_mute(user_screen_name)


    def tweet(self, tweet_text):
        self.update_status(tweet_text)

    def delete_tweet(self, tweet_id):
        deleted_tweet = self.get_status(tweet_id)
        
        self.destroy_status(tweet_id)


        
    def retweet_(self, tweet_id):
        self.retweet(tweet_id)

    def unretweet_(self, tweet_id):
        unretweeted_tweet = self.get_status(tweet_id)
        
        self.unretweet(tweet_id)
        
    def like_(self, tweet_id):
        self.create_favorite(tweet_id)

    def unlike(self, tweet_id):
        self.destroy_favorite(tweet_id)


    def reply(self, comment, tweet_from, tweet_id):
        # Reply should have screen name of who you want to reply to and the tweet you want to reply on
        self.update_status(f'{comment} @{tweet_from}', in_reply_to_status_id=tweet_id, 
            auto_populate_reply_metadata=True)
        
        # logging
        reply_to = self.get_status(tweet_id)
        

    def delete_reply(self, tweet_id):
        self.delete_tweet(tweet_id)

    def quoted_tweet(self, comment, tweet_from, tweet_id):
        # Quoted tweet is ordinary tweet with another tweet link and comment 
        tweet_url = f'https://twitter.com/{tweet_from}/status/{tweet_id}'
        
        self.update_status(f'{comment} \n {tweet_url}')


    def send_message(self,recipient_id,message_text):
        self.send_direct_message(recipient_id,message_text)

    def do(self, action):
        try:
            eval(f"self.{action.name}(**{action.params})")

        except:
            pass
            


